/*****************************************/
/* Tetris                                */
/* (C) 2000, akop karapetyan             */
/*                                       */
/* Class Cup                             */
/*   Actual game                         */
/*                                       */
/*****************************************/

import java.awt.Image;

import java.awt.Canvas;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;

public final class Cup extends Canvas implements KeyListener {
	
	private final int 
		m = 20,		/* number of rows in a cup */
		n = 10,		/* number of columns in a cup */
		p = 3;		/* width of the padding on each side */
	
	private final int 
		hOffset = 15,
		vOffset = 15;
	
	private int
		x, y,		/* column and row of current trinket */
		gx, gy;		/* column and row of the ghosted trinket */
	
	private int 
		angle,		/* Angle of current trinket (actual angle is value * 90 degrees) */
		current;	/* Type of current trinket */
	
	private boolean refresh;	/* Flag that specifies whether to repaint the cup */
	
	private int[][] cup = new int[m + p * 2][n + p * 2];	/* Array, holds the cup */
	
	private int[][] trinket;		/* holds current trinket */
	private int[][] ghost;			/* holds current ghost */
	
	private Color bgColor = new Color(64, 64, 64);
	private Game game;
	
	private Image blocks, logo, buffer;

	public Graphics bufg;
	
	/* Constructor */
	public Cup(Game p_game, Image[] gfx) {
		
		int
			width  = n * p_game.w + hOffset * 2 + 15,
			height = m * p_game.h + vOffset * 2 + 5;
		
		game = p_game;
		
		addKeyListener(this);
		setBackground(game.getBackground());		
		setSize(width, height);
		
		blocks = gfx[0];
		buffer = gfx[2];
		logo = gfx[4];

		bufg = buffer.getGraphics();
		
		newGame();
	}
	
	/* Inits the cup and sets up some things */
	public void newGame() {
		
		for (int i = 0; i < cup.length; i++)
			for (int j = 0; j < cup[0].length; j++)
				if ( (i < (m + p)) && ((j >= p) && (j < (n + p))) )
					cup[i][j] = 0;
				else
					cup[i][j] = 1;
		
		game.getNextTrinket();
		routineInit();
	}
	
	/* Called every time a trinket "drops", or at initialization */
	private void routineInit() {
		
		angle = 0;
		current = game.next;
		game.getNextTrinket();
		
		y = gy = 2;//p;
		x = gx = p + (n / 2);
		
		trinket = new Trinket(current, angle).get2DIntArray();
		if ( !collisionCheck(trinket, y, x) ) {
			
			game.event(game.GAMEOVER, 0);
			return;
		}
		
		game.event(game.NEXT, 0);
		
		setGhostToTrinket();
		refresh = true;
	}
	
	/*	Simply setting 'ghost = trinket' seems to cause odd problems with 
		Microsoft's VM resulting in corrupted new ghost arrays. Therefore 
		an explicit bit-by-bit assignment is necessary (same reason for 
		copying the cup bit-by-bit in dropRows). This problem doesn't seem 
		to appear under Netscape oddly enough */
	private void setGhostToTrinket() {
		
		int gM = trinket.length,
			gN = trinket[0].length;
		
		ghost = new int[gM][gN];
		
		for (int i = 0; i < gM; i++)
			for (int j = 0; j < gN; j++)
				ghost[i][j] = trinket[i][j];
	}
	
	/* Does a "bitwise" check to see whether a piece can
	   be placed at the specified coordinates 
			i.e. nothing is blocking it from being placed
			in the destination */
	private boolean collisionCheck( int[][] nT, int y, int x ) {
		
		for (int i = 0; i < nT.length; i++)
			for (int j = 0; j < nT[0].length; j++)
				if ( (nT[i][j] != 0) && (cup[ i + y ][ j + x ] != 0) )
					return false;
		
		return true;			
	}
	
	/* Takes the current trinket and "paints" it at the 
	   specified coordinates in the cup */
	private void overlay(int y, int x) {
		
		for (int i = 0; i < trinket.length; i++)
			for (int j = 0; j < trinket[0].length; j++)
				if ( trinket[i][j] != 0 )
					cup[i + y][j + x] = trinket[i][j];
	}
	
	/* Called whenever the applet needs to be redrawn */
	public void paint(Graphics g) {
		
		drawCup(g);
		refresh = true; /* Force repainting of the cup */
		clearCup(g);
		myPaint(g);
	}

	public void update(Graphics g) {

		paint(g);
	}

	/*	Clears the cup, so it can be repainted. Used in case
		some rows drop */
	private void clearCup(Graphics g) {
		
		bufg.setColor(bgColor);
		bufg.fillRect(hOffset, vOffset, game.w * n, game.h * m);
	}
	
	/* This is the actual rendering function */
	private void myPaint(Graphics g) {
		
		if (g == null)
			return;
		
		/* This part re-draws the cup */
		if (refresh) {
			
			for (int i = 0; i < m; i++)
				for (int j = 0; j < n; j++)
					if (cup[i + p][j + p] != 0)
						bufg.drawImage(blocks, 
									   (j * game.w) + hOffset, 
									   (i * game.h) + vOffset, 
									   ((j * game.w) + hOffset) + game.w, 
									   ((i * game.h) + vOffset) + game.h,
									   (cup[i + p][j + p] - 1) * game.w, 0, 
									   ((cup[i + p][j + p] - 1) * game.w) + game.w, 16, this);
			
			refresh = false;
		}
		
		if (game.gameStatus()) {
			/* This part erases the old trinket (ghost) */
			if (gy > -1)
				for (int i = 0; i < ghost.length; i++)
					for (int j = 0; j < ghost[0].length; j++)
						if ((ghost[i][j] != 0) && ((i + gy) > 2)) {
							bufg.setColor(bgColor);
							bufg.fillRect(((j + gx - p) * game.w) + hOffset, ((i + gy - p) * game.h) + vOffset, game.w, game.h);
						}
		
			/* This part draws the current trinket */
			for (int i = 0; i < trinket.length; i++)
				for (int j = 0; j < trinket[0].length; j++) 
					if ((trinket[i][j] != 0) && ((i + y) > 2))
						bufg.drawImage(blocks, 
						   ((j + x - p) * game.w) + hOffset, 
						   ((i + y - p) * game.h) + vOffset, 
						   (((j + x - p) * game.w) + hOffset) + game.w, 
						   (((i + y - p) * game.h) + vOffset) + game.h,
						   (trinket[i][j] - 1) * game.w, 0, 
						   ((trinket[i][j] - 1) * game.w) + game.w, 16, this);
		} else 
			bufg.drawImage(logo, 15, 143, this);
		
		/* Dbl-buffering */
		g.drawImage(buffer, 0, 0, this);
	}
	
	/*	Draws the cup window (3D effects n such) */
	private void drawCup(Graphics g) {
		
		int bWidth = 5;
		int boxLeft = hOffset - bWidth - 10,
			boxTop = vOffset - bWidth - 10,
			boxWide = (n * game.w) + (bWidth * 2) + 20,
			boxHigh = (m * game.h) + (bWidth * 2) + 20;
		
		bufg.setColor(new Color(32, 32, 32));
		for (int i = 0; i < 5; i++)
			bufg.drawRect(i, i, 190, 350);
		
		bufg.setColor(new Color(128, 128, 128));
		
		for (int i = 0; i < bWidth; i++)
			bufg.fill3DRect(boxLeft + i, boxTop + i, 
						boxWide - i * 2, boxHigh - i * 2, true);
		
		bufg.draw3DRect(hOffset - 1, vOffset - 1, 
					 (n * game.w) + 2, (m * game.h) + 2, false);		
	}
	
	/* Drops full rows (if any) and returns the number of dropped rows */
	private int dropFullRows() {
		boolean broke;
		int dropped = 0;
		
		/*	To ensure normal operation timer must be paused before
			entering this function */
		for (int i = (m + p - 1); i >= p; i--) {
			broke = false;
			
			for (int j = p; j < (n + p); j++) {
				if ( cup[i][j] == 0 ) {
					broke = true;
					break;
				}
			}
			
			/* The row is full, needs to be dropped */
			if (!broke) {
				
				for (int k = i; k >= p; k--)
					for (int l = p; l < (n + p); l++)
						cup[k][l] = cup[ k - 1 ][l];
				
				i++; /* Rows have been shifted, current row must be re-evaluated */
				dropped++;
				refresh = true;
			}
		}
		
		if (refresh) {
			
			clearCup(getGraphics());
			paint(getGraphics());
		}
		
		return dropped;
	}
	
	/* Performs an action */
	public void act(int how) {
		boolean posChange = false;
		
		switch (how) {
			case 0: {
				if ( collisionCheck(trinket, y, x - 1) ) {
					setGhostToTrinket();
					gx = x;
					gy = y;
					x--;
					
					posChange = true;
				}
				
				break;
			}
			case 1: {
				if ( collisionCheck(trinket, y, x + 1) ) {
					setGhostToTrinket();
					gx = x;
					gy = y;
					x++;
					
					posChange = true;
				}
				
				break;
			}
			case 2: {
				if ( collisionCheck(trinket, y + 1, x) ) {
					/* Further progression down is possible */
					setGhostToTrinket();
					gy = y;
					gx = x;
					y++;
					
					posChange = true;
				} else {
					/* No further progression is possible */
					overlay(y, x);
					
					routineInit();
					gy = -1;
					
					refresh = false;
					myPaint(getGraphics());
					
					game.pauseTimer();					
					game.event(game.DROP, dropFullRows());
					game.resumeTimer();
				}
				
				break;
			}
			case 3: {
				int[][] temp = new Trinket(current, angle + 1).get2DIntArray();
				
				if (collisionCheck(temp, y, x)) {
					gx = x;
					gy = y;
					setGhostToTrinket();
					
					trinket = new Trinket(current, angle + 1).get2DIntArray();
					
					if ( ++angle > 3 ) 
						angle = 0;
					
					posChange = true;
				}
				
				break;
			}
			case 4: {
				game.pauseTimer();
				
				for (int i = y + 1; i < (m + p + 1); i++)
					if (! collisionCheck(trinket, i, x)) {
						/* No further progression is possible */
						y = i - 1;
						overlay(y, x);
						
						routineInit();
						gy = -1;
						
						refresh = true;						
						clearCup(getGraphics());
						myPaint(getGraphics());
						refresh = false;						
						
						game.event(game.DROP, dropFullRows());
						break;
					}
				
				game.resumeTimer();
				break;
			}
		}
		
		if (posChange)
			myPaint(getGraphics());
		
	}
	
	/* Occurs whenever a key is pressed */
	public void keyPressed(KeyEvent e) {
		
		if (game.gameStatus())
			
			switch (e.getKeyCode()) {
				case KeyEvent.VK_LEFT: {
					act(0);
					
					break;
				}
				case KeyEvent.VK_RIGHT: {
					act(1);
					
					break;
				}
				case KeyEvent.VK_DOWN: {
					act(2);
					
					break;
				}
				case KeyEvent.VK_SPACE: {
					act(3);
					
					break;
				}
				case KeyEvent.VK_D: {
					act(4);
					
					break;
				}
			}
		else

			if (e.getKeyCode() == KeyEvent.VK_SPACE)
				game.event(game.GAMESTART, 0);
	}
	
	/* KeyListener interface requires this */
	public void keyReleased(KeyEvent e) {}	
	public void keyTyped(KeyEvent e) {}
}
